"""
    .. Copyright:
        Copyright 2019 Wirepas Ltd licensed under Apache License, Version 2.0
        See file LICENSE for full license details.
"""

__author__ = "Wirepas Ltd"
__author_email__ = "opensource@wirepas.com"
__classifiers__ = [
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: Apache Software License",
    "Topic :: Software Development :: Libraries",
    "Programming Language :: Python :: 3",
]
__copyright__ = "2019 Wirepas Ltd"
__description__ = "Wirepas gateway transport service that connects the local dbus to a remote MQTT broker."
__license__ = "Apache-2"
__pkg_name__ = "wirepas_gateway"
__title__ = "Wirepas Gateway Transport Service"
__url__ = "https://github.com/wirepas/gateway"
__version__ = "1.4.5rc1"
__keywords__ = "wirepas connectivity iot mesh"
